### Persefone.jl - a model of agricultural landscapes and ecosystems in Europe.
###
### Functions for switchable crop models.
###

using StatsBase: mode

"""
    initcropmodels(cropmodels, cropdirectory)

Initialise the crop models given as a comma-delimited string
(e.g. "almass,aquacrop"). The crop model parameters are read from the
`cropdirectories`, also a comma-delimited string. Returns the crop
types available in the simulation.
"""
function initcropmodels(cropmodels::AbstractString,
                        cropdirectories::AbstractString)
    models = string.(split(cropmodels, ","))
    dirs = string.(split(cropdirectories, ","))
    if length(models) != length(dirs)
        error("Different number of cropmodels and cropdirectories.\n" *
              "cropmodels = $cropmodels\ncropdirectories = $cropdirectories")
    end

    allcrops = Dict{String, AbstractCropType}()
    for (i, cropmodel) in enumerate(models)
        cropdirectory = dirs[i]
        if cropmodel == "almass"
            crops = ALMaSS.readcropparameters(cropdirectory)
        elseif cropmodel == "simple"
            crops_almass = ALMaSS.readcropparameters(cropdirectory)
            crops = Dict(name => SimpleCrop.CropType(ct.name, ct.group, ct.minsowdate, ct.maxsowdate)
                         for (name, ct) in crops_almass)
        elseif cropmodel == "aquacrop"
            crops = AquaCropWrapper.readcropparameters(cropdirectory)
        else
            error("initcropmodel: no implementation for crop model '$cropmodel'")
        end
        merge!(allcrops, crops)
    end
    return allcrops
end

"""
    initfields!(model)

Initialise the farm plots in the simulation model.
"""
function initfields!(model::SimulationModel)
    # TODO: we take the soiltype from the first pixel in the farmplot,
    # then later we fix the soiltypes of the farmplots and cropstates
    # to be the most common soiltype of a farmplot.  It would be
    # better to first collect all the data for pixels belonging
    # together, and then create the farmplots and cropstates with the
    # correct soiltype directly
    convertid = Dict{Int64,Int64}()
    width, height = size(model.landscape)
    for x in 1:width
        for y in 1:height
            # for each pixel, we need to extract the field ID given by the map input
            # file, and convert it into an internal object ID for `model.farmplots`
            rawid = model.landscape[x,y].fieldid
            (ismissing(rawid)) && continue
            if rawid in keys(convertid)
                objectid = convertid[rawid]
                model.landscape[x,y].fieldid = objectid
                push!(model.farmplots[objectid].pixels, (x,y))
            else
                soiltype = model.landscape[x,y].soiltype
                cropstate = makecropstate(model.crops["natural grass"], model, soiltype)
                fp = FarmPlot(length(model.farmplots) + 1, [(x, y)], -1, soiltype, cropstate)
                push!(model.farmplots, fp)
                model.landscape[x,y].fieldid = fp.id
                convertid[rawid] = fp.id
            end
        end
    end

    # Adjust farmplot soil type to most common soil type of its
    # pixels, and set the cropstate soil type to this as well.
    for fp in model.farmplots
        soiltype = mode(map(p -> model.landscape[p[1], p[2]].soiltype, fp.pixels))
        setsoiltype!(fp, soiltype)
    end

    @info "Initialised $(length(model.farmplots)) farm plots."
end
